import math
from typing import ClassVar

from gmpy2 import mpz

__all__ = ["EncoderBase"]


class EncoderBase:
    alphabet: ClassVar[str]
    base: ClassVar[mpz]
    zero: ClassVar[str]
    lookup: ClassVar[dict[str, mpz]]

    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        cls.base = mpz(len(cls.alphabet))
        cls.zero = cls.alphabet[0]
        cls.lookup = {c: mpz(i) for i, c in enumerate(cls.alphabet)}

    @classmethod
    def mpz_to_base(cls, number: mpz) -> str:
        result: list[str] = []
        while number:
            number, remainder = divmod(number, cls.base)
            result.append(cls.alphabet[int(remainder)])

        return "".join(reversed(result))

    @classmethod
    def base_to_mpz(cls, data: str) -> mpz:
        number = mpz(0)
        for digit in data:
            number = number * cls.base + cls.lookup[digit]

        return number

    @classmethod
    def encode(cls, data: bytes) -> str:
        if not data:
            return ""

        count = 0
        for i in data:
            if i == 0:
                count += 1
            else:
                break

        number = mpz.from_bytes(data, byteorder="big", signed=False)
        if number == 0:
            return cls.zero * count

        encoded = cls.mpz_to_base(number)
        return cls.zero * count + encoded

    @classmethod
    def decode(cls, data: str) -> bytes:
        stripped = data.lstrip(cls.zero)
        count = len(data) - len(stripped)
        if stripped:
            number = cls.base_to_mpz(stripped)
            byte_len = math.ceil(number.bit_length() / 8)
            decoded = number.to_bytes(byte_len, byteorder="big", signed=False)
        else:
            decoded = b""

        return b"\x00" * count + decoded



import base64
import string
from typing import ClassVar

from .encoder import Encoder
from .encoder_base import EncoderBase

__all__ = [
    "Base10Encoder",
    "Base26Encoder",
    "Base31Encoder",
    "Base32Encoder",
    "Base36Encoder",
    "Base64Encoder",
    "Base94Encoder",
    "decode_bytes",
    "encode_str",
]


def encode_str(data: str) -> bytes:
    return data.encode(encoding="utf-8", errors="strict")


def decode_bytes(data: bytes) -> str:
    return data.decode(encoding="utf-8", errors="strict")


class Base10Encoder(EncoderBase, Encoder):
    alphabet: ClassVar[str] = string.digits


class Base26Encoder(EncoderBase, Encoder):
    alphabet: ClassVar[str] = string.ascii_uppercase


class Base31Encoder(EncoderBase, Encoder):
    alphabet: ClassVar[str] = "АБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЬЫЭЮЯ"


class Base32Encoder(Encoder):
    @classmethod
    def encode(cls, data: bytes) -> str:
        return decode_bytes(base64.b32encode(data))

    @classmethod
    def decode(cls, data: str) -> bytes:
        return base64.b32decode(data)


class Base36Encoder(EncoderBase, Encoder):
    alphabet: ClassVar[str] = string.digits + string.ascii_uppercase


class Base64Encoder(Encoder):
    @classmethod
    def encode(cls, data: bytes) -> str:
        return decode_bytes(base64.b64encode(data))

    @classmethod
    def decode(cls, data: str) -> bytes:
        return base64.b64decode(data)


class Base94Encoder(EncoderBase, Encoder):
    alphabet: ClassVar[str] = "".join([chr(i) for i in range(33, 127)])



from typing import Protocol

__all__ = ["Encoder"]


class Encoder(Protocol):
    @classmethod
    def encode(cls, data: bytes) -> str: ...

    @classmethod
    def decode(cls, data: str) -> bytes: ...



from typing import TYPE_CHECKING

import click
from click_aliases import ClickAliasedGroup

from rtty_soda.cli import CliOptions, CliReader, CliTypes, CliWriter
from rtty_soda.formatters import FixedFormatter
from rtty_soda.services import EncodingService, EncryptionService, KeyService
from rtty_soda.totp import GoogleAuthenticator

if TYPE_CHECKING:
    from pathlib import Path

__all__ = ["cli"]


CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(cls=ClickAliasedGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option(package_name="rtty-soda")
def cli() -> None:
    pass


@cli.command()  # pyright: ignore[reportAny]
@CliOptions.key_passphrase
@CliOptions.short_key_encoding
@CliOptions.output_file
@CliOptions.group_len
@CliOptions.line_len
@CliOptions.padding
@CliOptions.verbose
def genkey_cmd(
    key_passphrase: str | None,
    encoding: str,
    output_file: Path | None,
    group_len: int,
    line_len: int,
    padding: int,
    verbose: bool,
) -> None:
    """Generate private/secret key."""
    formatter = FixedFormatter(group_len, line_len, padding)
    writer = CliWriter(output_file)
    service = KeyService(encoding, formatter, writer, verbose)
    service.genkey(passphrase=key_passphrase)


@cli.command()  # pyright: ignore[reportAny]
@click.argument("private_key_file", type=CliTypes.IN_PATH)
@CliOptions.key_passphrase
@CliOptions.short_key_encoding
@CliOptions.output_file
@CliOptions.group_len
@CliOptions.line_len
@CliOptions.padding
@CliOptions.verbose
def pubkey_cmd(
    private_key_file: Path,
    key_passphrase: str | None,
    encoding: str,
    output_file: Path | None,
    group_len: int,
    line_len: int,
    padding: int,
    verbose: bool,
) -> None:
    """Get public key."""
    formatter = FixedFormatter(group_len, line_len, padding)
    writer = CliWriter(output_file)
    service = KeyService(encoding, formatter, writer, verbose)
    private_key = CliReader(private_key_file)
    service.pubkey(private_key=private_key, passphrase=key_passphrase)


@cli.command()  # pyright: ignore[reportAny]
@click.argument("password_file", type=CliTypes.IN_PATH)
@CliOptions.key_passphrase
@CliOptions.short_key_encoding
@CliOptions.short_kdf_profile
@CliOptions.output_file
@CliOptions.group_len
@CliOptions.line_len
@CliOptions.padding
@CliOptions.verbose
def kdf_cmd(
    password_file: Path,
    key_passphrase: str | None,
    encoding: str,
    profile: str,
    output_file: Path | None,
    group_len: int,
    line_len: int,
    padding: int,
    verbose: bool,
) -> None:
    """Key derivation function."""
    formatter = FixedFormatter(group_len, line_len, padding)
    writer = CliWriter(output_file)
    service = KeyService(encoding, formatter, writer, verbose)
    password = CliReader(password_file)
    service.kdf(password=password, kdf_profile=profile, passphrase=key_passphrase)


@cli.command(aliases=["e"])  # pyright: ignore[reportAny]
@click.argument("private_key_file", type=CliTypes.IN_PATH)
@click.argument("public_key_file", type=CliTypes.IN_PATH)
@click.argument("message_file", type=CliTypes.IN_PATH)
@CliOptions.text
@CliOptions.key_passphrase
@CliOptions.key_encoding
@CliOptions.data_encoding
@CliOptions.compression
@CliOptions.output_file
@CliOptions.group_len
@CliOptions.line_len
@CliOptions.padding
@CliOptions.verbose
def encrypt_public_cmd(
    private_key_file: Path,
    public_key_file: Path,
    message_file: Path,
    text: bool,
    key_passphrase: str | None,
    key_encoding: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
    group_len: int,
    line_len: int,
    padding: int,
    verbose: bool,
) -> None:
    """Encrypt message (public)."""
    formatter = FixedFormatter(group_len, line_len, padding)
    writer = CliWriter(output_file)
    service = EncryptionService(
        text_mode=text,
        key_encoding=key_encoding,
        data_encoding=data_encoding,
        compression=compression,
        formatter=formatter,
        writer=writer,
        verbose=verbose,
    )
    private_key = CliReader(private_key_file)
    public_key = CliReader(public_key_file)
    message = CliReader(message_file)
    service.encrypt_public(
        private_key=private_key,
        passphrase=key_passphrase,
        public_key=public_key,
        message=message,
    )


@cli.command(aliases=["es"])  # pyright: ignore[reportAny]
@click.argument("key_file", type=CliTypes.IN_PATH)
@click.argument("message_file", type=CliTypes.IN_PATH)
@CliOptions.text
@CliOptions.key_passphrase
@CliOptions.key_encoding
@CliOptions.data_encoding
@CliOptions.compression
@CliOptions.output_file
@CliOptions.group_len
@CliOptions.line_len
@CliOptions.padding
@CliOptions.verbose
def encrypt_secret_cmd(
    key_file: Path,
    message_file: Path,
    text: bool,
    key_passphrase: str | None,
    key_encoding: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
    group_len: int,
    line_len: int,
    padding: int,
    verbose: bool,
) -> None:
    """Encrypt message (secret)."""
    formatter = FixedFormatter(group_len, line_len, padding)
    writer = CliWriter(output_file)
    service = EncryptionService(
        text_mode=text,
        key_encoding=key_encoding,
        data_encoding=data_encoding,
        compression=compression,
        formatter=formatter,
        writer=writer,
        verbose=verbose,
    )
    key = CliReader(key_file)
    message = CliReader(message_file)
    service.encrypt_secret(key=key, passphrase=key_passphrase, message=message)


@cli.command(aliases=["ep"])  # pyright: ignore[reportAny]
@click.argument("password_file", type=CliTypes.IN_PATH)
@click.argument("message_file", type=CliTypes.IN_PATH)
@CliOptions.text
@CliOptions.kdf_profile
@CliOptions.data_encoding
@CliOptions.compression
@CliOptions.output_file
@CliOptions.group_len
@CliOptions.line_len
@CliOptions.padding
@CliOptions.verbose
def encrypt_password_cmd(
    password_file: Path,
    message_file: Path,
    text: bool,
    kdf_profile: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
    group_len: int,
    line_len: int,
    padding: int,
    verbose: bool,
) -> None:
    """Encrypt message (password)."""
    formatter = FixedFormatter(group_len, line_len, padding)
    writer = CliWriter(output_file)
    service = EncryptionService(
        text_mode=text,
        key_encoding="binary",
        data_encoding=data_encoding,
        compression=compression,
        formatter=formatter,
        writer=writer,
        verbose=verbose,
    )
    pw = CliReader(password_file)
    message = CliReader(message_file)
    service.encrypt_password(password=pw, message=message, kdf_profile=kdf_profile)


@cli.command(aliases=["d"])  # pyright: ignore[reportAny]
@click.argument("private_key_file", type=CliTypes.IN_PATH)
@click.argument("public_key_file", type=CliTypes.IN_PATH)
@click.argument("message_file", type=CliTypes.IN_PATH)
@CliOptions.text
@CliOptions.key_passphrase
@CliOptions.key_encoding
@CliOptions.data_encoding
@CliOptions.compression
@CliOptions.output_file
def decrypt_public_cmd(
    private_key_file: Path,
    public_key_file: Path,
    message_file: Path,
    text: bool,
    key_passphrase: str | None,
    key_encoding: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
) -> None:
    """Decrypt message (public)."""
    writer = CliWriter(output_file)
    service = EncryptionService(
        text_mode=text,
        key_encoding=key_encoding,
        data_encoding=data_encoding,
        compression=compression,
        formatter=None,
        writer=writer,
        verbose=False,
    )
    private_key = CliReader(private_key_file)
    public_key = CliReader(public_key_file)
    message = CliReader(message_file)
    service.decrypt_public(
        private_key=private_key,
        passphrase=key_passphrase,
        public_key=public_key,
        message=message,
    )


@cli.command(aliases=["ds"])  # pyright: ignore[reportAny]
@click.argument("key_file", type=CliTypes.IN_PATH)
@click.argument("message_file", type=CliTypes.IN_PATH)
@CliOptions.text
@CliOptions.key_passphrase
@CliOptions.key_encoding
@CliOptions.data_encoding
@CliOptions.compression
@CliOptions.output_file
def decrypt_secret_cmd(
    key_file: Path,
    message_file: Path,
    text: bool,
    key_passphrase: str | None,
    key_encoding: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
) -> None:
    """Decrypt message (secret)."""
    writer = CliWriter(output_file)
    service = EncryptionService(
        text_mode=text,
        key_encoding=key_encoding,
        data_encoding=data_encoding,
        compression=compression,
        formatter=None,
        writer=writer,
        verbose=False,
    )
    key = CliReader(key_file)
    message = CliReader(message_file)
    service.decrypt_secret(key=key, passphrase=key_passphrase, message=message)


@cli.command(aliases=["dp"])  # pyright: ignore[reportAny]
@click.argument("password_file", type=CliTypes.IN_PATH)
@click.argument("message_file", type=CliTypes.IN_PATH)
@CliOptions.text
@CliOptions.kdf_profile
@CliOptions.data_encoding
@CliOptions.compression
@CliOptions.output_file
def decrypt_password_cmd(
    password_file: Path,
    message_file: Path,
    text: bool,
    kdf_profile: str,
    data_encoding: str,
    compression: str,
    output_file: Path | None,
) -> None:
    """Decrypt message (password)."""
    writer = CliWriter(output_file)
    service = EncryptionService(
        text_mode=text,
        key_encoding="binary",
        data_encoding=data_encoding,
        compression=compression,
        formatter=None,
        writer=writer,
        verbose=False,
    )
    pw = CliReader(password_file)
    message = CliReader(message_file)
    service.decrypt_password(password=pw, message=message, kdf_profile=kdf_profile)


@cli.command()  # pyright: ignore[reportAny]
@click.argument("in_encoding", type=CliTypes.ENCODING, metavar="IN_ENCODING")
@click.argument("out_encoding", type=CliTypes.ENCODING, metavar="OUT_ENCODING")
@click.argument("file", type=CliTypes.IN_PATH)
@CliOptions.output_file
@CliOptions.group_len
@CliOptions.line_len
@CliOptions.padding
@CliOptions.verbose
def encode_cmd(
    in_encoding: str,
    out_encoding: str,
    file: Path,
    output_file: Path | None,
    group_len: int,
    line_len: int,
    padding: int,
    verbose: bool,
) -> None:
    """Encode file.

    See `soda encodings` for available encodings.
    """
    writer = CliWriter(output_file)
    formatter = FixedFormatter(group_len, line_len, padding)
    service = EncodingService(
        in_encoding=in_encoding,
        out_encoding=out_encoding,
        formatter=formatter,
        writer=writer,
        verbose=verbose,
    )
    data = CliReader(file)
    service.encode(data)


@cli.command(aliases=["ga"])  # pyright: ignore[reportAny]
@click.argument("key_file", type=CliTypes.IN_PATH)
def google_auth_cmd(key_file: Path) -> None:
    """Google Authenticator TOTP.

    Key must be Base32-encoded.

    When importing the key into Google Authenticator, strip the trailing '=' characters.
    """
    key = CliReader(key_file)
    totp = GoogleAuthenticator(key)
    click.echo(f"{totp.get_code()} (expires in {totp.get_remaining_seconds()}s)")


@cli.command()  # pyright: ignore[reportAny]
def encodings_cmd() -> None:
    """List supported encodings."""
    help_text = """
    base10 (Decimal)
    base26 (Latin)
    base31 (Cyrillic)
    base32 (RFC 4648)
    base36 (Latin with digits)
    base64 (RFC 4648)
    base94 (ASCII printable)
    binary (Raw bytes)

    Use binary or base64 for files larger than 1 MiB.
    """
    click.echo(help_text)


@cli.command()  # pyright: ignore[reportAny]
def compression_cmd() -> None:
    """List supported compression libs."""
    help_text = """
    brotli (Brotli, best ratio, default)
    zstd (Zstandard, fast, good ratio)
    zlib (Deflate, used by gzip)
    bz2 (Bzip2, used in tar.bz2)
    lzma (LZMA, good ratio on large data)
    raw (No compression, improves security)
    """
    click.echo(help_text)


@cli.command()  # pyright: ignore[reportAny]
def kdf_profiles_cmd() -> None:
    """List supported KDF profiles."""
    help_text = """
    interactive (64 MiB, 2 passes)
    moderate (256 MiB, 3 passes)
    sensitive (1 GiB, 4 passes)
    """
    click.echo(help_text)


if __name__ == "__main__":
    cli()



from .concrete import (
    Base10Encoder,
    Base26Encoder,
    Base31Encoder,
    Base32Encoder,
    Base36Encoder,
    Base64Encoder,
    Base94Encoder,
    decode_bytes,
    encode_str,
)
from .encoder import Encoder
from .encoder_base import EncoderBase
from .scsu import scsu_decode, scsu_encode

__all__ = [
    "ENCODERS",
    "Base10Encoder",
    "Base26Encoder",
    "Base31Encoder",
    "Base32Encoder",
    "Base36Encoder",
    "Base64Encoder",
    "Base94Encoder",
    "Encoder",
    "EncoderBase",
    "decode_bytes",
    "encode_str",
    "scsu_decode",
    "scsu_encode",
]

ENCODERS: dict[str, Encoder | None] = {
    "base10": Base10Encoder(),
    "base26": Base26Encoder(),
    "base31": Base31Encoder(),
    "base32": Base32Encoder(),
    "base36": Base36Encoder(),
    "base64": Base64Encoder(),
    "base94": Base94Encoder(),
    "binary": None,
}



from typing import TYPE_CHECKING

from nacl.public import PrivateKey

from rtty_soda.cryptography import kdf, secret
from rtty_soda.encoders import ENCODERS, encode_str

from .service import Service

if TYPE_CHECKING:
    from rtty_soda.cryptography.secret import SecretKeyOp
    from rtty_soda.formatters import Formatter
    from rtty_soda.interfaces import Reader, Writer

__all__ = ["KeyService"]


class KeyService(Service):
    def __init__(
        self, encoding: str, formatter: Formatter, writer: Writer, verbose: bool
    ) -> None:
        super().__init__(formatter, writer, verbose)
        self.encoder = ENCODERS.get(encoding)

    def genkey(self, passphrase: str | None) -> None:
        priv_bytes = bytes(PrivateKey.generate())
        protected = self.protect(private_key=priv_bytes, passphrase=passphrase)
        self.write_output(protected, self.encoder)

    def pubkey(self, private_key: Reader, passphrase: str | None) -> None:
        priv_bytes = self.read_input(private_key, self.encoder)
        unprotected = self.unprotect(private_key=priv_bytes, passphrase=passphrase)
        priv_key = PrivateKey(unprotected)
        public_key = bytes(priv_key.public_key)
        self.write_output(public_key, self.encoder)

    def kdf(self, password: Reader, kdf_profile: str, passphrase: str | None) -> None:
        key_bytes = self.derive_key(password, kdf_profile)
        protected = self.protect(private_key=key_bytes, passphrase=passphrase)
        self.write_output(protected, self.encoder)

    @staticmethod
    def derive_key(password: str | Reader, kdf_profile: str) -> bytes:
        prof = kdf.KDF_PROFILES[kdf_profile]
        pw_str = password if isinstance(password, str) else password.read_str()
        pw_bytes = encode_str(pw_str.strip())
        return kdf.kdf(password=pw_bytes, profile=prof)

    @staticmethod
    def passphrase_flow(
        private_key: bytes, passphrase: str | None, operation: SecretKeyOp
    ) -> bytes:
        if passphrase is None:
            return private_key

        skey = KeyService.derive_key(passphrase, "sensitive")
        return operation(key=skey, data=private_key)

    @staticmethod
    def protect(private_key: bytes, passphrase: str | None) -> bytes:
        return KeyService.passphrase_flow(
            private_key=private_key, passphrase=passphrase, operation=secret.encrypt
        )

    @staticmethod
    def unprotect(private_key: bytes, passphrase: str | None) -> bytes:
        return KeyService.passphrase_flow(
            private_key=private_key, passphrase=passphrase, operation=secret.decrypt
        )
